# abcpmc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# abcpmc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with abcpmc.  If not, see <http://www.gnu.org/licenses/>.

import itertools

try:
    from mpi4py import MPI
    MPI = MPI
except ImportError:
    MPI = None

__all__ = ["MpiPool", "mpiBCast"]

class MpiPool(object):
    def __init__(self, mapFunction=map):
        self.rank = MPI.COMM_WORLD.Get_rank()
        self.size = MPI.COMM_WORLD.Get_size()
        self.mapFunction = mapFunction
    
    def map(self, function, sequence):
        (rank,size) = (MPI.COMM_WORLD.Get_rank(),MPI.COMM_WORLD.Get_size())
        sequence = mpiBCast(sequence)
        mergedList = _mergeList(MPI.COMM_WORLD.allgather(
                                                  self.mapFunction(function, _splitList(sequence,size)[rank])))
        return mergedList
    
    def isMaster(self):
        """
        Returns true if the rank is 0
        """
        return (self.rank==0)
    
def mpiBCast(value):
    """
    Mpi bcasts the value and returns the value from the master (rank = 0).
    """
    return MPI.COMM_WORLD.bcast(value)

def _splitList(list, n):
    blockLen = len(list) / float(n)
    return [list[int(round(blockLen * i)): int(round(blockLen * (i + 1)))] for i in range(n)]    

def _mergeList(lists):
    return list(itertools.chain(*lists))






# abcpmc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# abcpmc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with abcpmc.  If not, see <http://www.gnu.org/licenses/>.
'''
Created on Jan 19, 2015

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

import numpy as np

class EpsProposal(object):
    
    def __init__(self, T):
        self.T = T
        self.reset()
        
    def __iter__(self):
        return self
    
    def __next__(self):
        return self.next()
    
    def next(self):
        if(self.t>=self.T):
            raise StopIteration()
        
        eps_val = self(self.t)
        self.t += 1
        return eps_val
    
    def reset(self):
        self.t = 0
        
class ListEps(EpsProposal):
    
    def __init__(self, T, eps_vals):
        super(ListEps, self).__init__(T)
        self.eps_vals = eps_vals
    
    def __call__(self, t):
        return self.eps_vals[t]

class ConstEps(EpsProposal):
    """
    Constant threshold. Can be used to apply alpha-percentile threshold decrease
    :param eps: epsilon value
    """
    
    def __init__(self, T, eps):
        super(ConstEps, self).__init__(T)
        self.eps = eps
        
    def __call__(self, t):
        return self.eps

class LinearEps(EpsProposal):
    """
    Linearly decreasing threshold
    
    :param max: epsilon at t=0
    :param min: epsilon at t=T
    :param T: number of iterations
    """
    
    def __init__(self, T, max, min):
        super(LinearEps, self).__init__(T)
        self.eps_vals = np.linspace(max, min, T)

    def __call__(self, t):
        return self.eps_vals[t]
        
class LinearConstEps(EpsProposal):
    """
    Linearly decreasing threshold until T1, then constant until T2
    
    :param max: epsilon at t=0
    :param min: epsilon at t=T
    :param T1: number of iterations for decrease
    :param T2: number of iterations for constant behavior
    """
    
    def __init__(self, max, min, T1, T2):
        super(LinearConstEps, self).__init__(T1+T2)
        self.eps_vals = np.r_[np.linspace(max, min, T1), [min]*T2]

    def __call__(self, t):
        return self.eps_vals[t]
        
class ExponentialEps(EpsProposal):
    """
    Exponentially decreasing threshold
    
    :param max: epsilon at t=0
    :param min: epsilon at t=T
    :param T: number of iterations
    """
    
    def __init__(self, T, max, min):
        super(ExponentialEps, self).__init__(T)
        self.eps_vals = np.logspace(np.log10(max), np.log10(min), T)

    def __call__(self, t):
        return self.eps_vals[t]
    
class ExponentialConstEps(EpsProposal):
    def __init__(self, max, min, T1, T2):
        super(ExponentialConstEps, self).__init__(T1+T2)
        self.eps_vals = np.r_[np.logspace(np.log10(max), np.log10(min), T1), [min]*T2]

    def __call__(self, t):
        return self.eps_vals[t]




# abcpmc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# abcpmc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with abcpmc.  If not, see <http://www.gnu.org/licenses/>.

__author__ = 'Joel Akeret'
__email__ = 'jakeret@phys.ethz.ch'
__version__ = '0.1.2'
__credits__ = 'ETH Zurich, Institute for Astronomy'

from abcpmc.sampler import *
from abcpmc.threshold import *



# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

"""
Tests for `abcpmc` module.
"""
from __future__ import print_function, division, absolute_import, unicode_literals

import abcpmc
import pytest
import numpy as np
from scipy import stats
from mock.mock import Mock


class TestTophatPrior(object):

    def test_tophat_univariate(self):
        min=1
        max=5
        with pytest.raises(AssertionError):
            abcpmc.TophatPrior(max, min)

        prior = abcpmc.TophatPrior(min, max)
        vals = prior()
        assert vals>min
        assert vals<max
        
        assert prior(theta=0)   == 0
        assert prior(theta=min) == 1
        assert prior(theta=2)   == 1
        assert prior(theta=max) == 0
        assert prior(theta=6)   == 0
        
    def test_tophat_multivariate(self):
        min=[1,2]
        max=[5, 6]
        with pytest.raises(AssertionError):
            abcpmc.TophatPrior(max, min)

        prior = abcpmc.TophatPrior(min, max)
        vals = prior()
        assert np.all(vals>min)
        assert np.all(vals<=max)
        
        assert prior(theta=[0, 0]) == 0
        assert prior(theta=[1, 2]) == 1
        assert prior(theta=[2, 3]) == 1
        assert prior(theta=[5, 5]) == 0
        assert prior(theta=[6, 6]) == 0

class TestGaussianPrior(object):
    
    def test_normal(self):
        prior = abcpmc.GaussianPrior([0,0], [[1,0],[0,1]])
        
        rngs = np.array([prior() for _ in range(10000)])
        assert len(rngs.shape) == 2
        D, p = stats.kstest(rngs[:,0], "norm")
        
        assert D < 0.015
        
    def test_pdf(self):
        try:
            from scipy.stats import multivariate_normal
        except ImportError:
            pytest.skip("Scipy.stats.multivariate_normal is not available")
            
        prior = abcpmc.GaussianPrior([0,0], [[1,0],[0,1]])
        
        theta = prior([0,0])
        assert np.allclose(theta, 0.15915, 1e-4)

class TestRejectionSamplingWrapperWrapper(object):
    
    def test_new_particle(self):
        eps = 1
        prior = lambda : 1
        thetai = 1
        postfn = lambda theta: thetai
        p = 0.5
        dist = lambda x,y: p
        Y = None
        wrapper = abcpmc.sampler._RejectionSamplingWrapper(eps, prior, postfn, dist, Y)
        rthetai, rp, cnt = wrapper(0)
        assert thetai == rthetai
        assert p == rp
        assert cnt == 1
        
#         swrapper = pickle.dumps(wrapper)
#         uppwrapper = pickle.loads(swrapper)

    def test_new_particle_multidist(self):
        eps = 1.
        threshold = [eps, eps]
        prior = lambda : 1
        thetai = 1
        postfn = lambda theta: thetai
        dist = Mock()
        
        distances = [[eps*2, eps*2], 
                    [eps/2, eps*2],
                    [eps*2, eps/2],
                    [eps, eps]]
        
        dist.side_effect = distances
        Y = None
        wrapper = abcpmc.sampler._RejectionSamplingWrapper(threshold, prior, postfn, dist, Y)
        _, _, cnt = wrapper(0)
        assert cnt == len(distances)


class TestWeightWrapper(object):

    def test_compute_weights(self):
        try:
            from scipy.stats import multivariate_normal
        except ImportError:
            pytest.skip("Scipy.stats.multivariate_normal is not available")
        
        
        prior = lambda theta: 1
        samples = [1] * 10
        weights = [1/len(samples)] * len(samples)
        sigma = 1
        
        wrapper = abcpmc.sampler._WeightWrapper(prior, sigma, weights, samples)
        rweight = wrapper(theta=0)
        
        assert rweight is not None
        assert rweight > 0.0
        
def test_weighted_cov():
    N = 4
    values = np.eye(N)
    weights = np.ones(N)
    wcov = abcpmc.weighted_cov(values, weights)
    
    assert np.all(wcov == np.cov(values))
    
class TestParticleProposal(object):
    
    def test_propose(self):
        eps = 1
        thetai = 1
        postfn = lambda theta: thetai
        p = 0.5
        dist = lambda x,y: p
        Y = None
        sampler = abcpmc.Sampler(2, Y, postfn, dist)
        
        thetas = np.array([[0.5], [1]])
        weights = np.array([0.75, 0.25])
        pool = abcpmc.sampler.PoolSpec(1, eps, 1, thetas, None, weights)
        
        wrapper = abcpmc.sampler.ParticleProposal(sampler, eps, pool, {})
        
        sigma = 0.25
        assert wrapper._get_sigma(None) == sigma
        
        rthetai, rp, cnt = wrapper(0)
        assert rthetai > (thetai - 5*sigma) and rthetai < (thetai + 5*sigma)
        assert p == rp
        assert cnt == 1
    
    def test_propose_multidist(self):
        eps = 1.
        threshold = [eps, eps]
        thetai = 1
        postfn = lambda theta: thetai
        dist = Mock()
        
        distances = [[eps*2, eps*2], 
                    [eps/2, eps*2],
                    [eps*2, eps/2],
                    [eps, eps]]
        
        dist.side_effect = distances
        Y = None
        sampler = abcpmc.Sampler(2, Y, postfn, dist)
        
        thetas = np.array([[0.5], [1]])
        weights = np.array([0.75, 0.25])
        pool = abcpmc.sampler.PoolSpec(1, threshold, 1, thetas, None, weights)
        
        wrapper = abcpmc.sampler.ParticleProposal(sampler, threshold, pool, {})
        
        _, _, cnt = wrapper(0)
        assert cnt == len(distances)
    
class TestOLCMParticleProposal(object):
    
    def test_get_sigma(self):
        eps = 1.1
        thetai = 1
        postfn = lambda theta: thetai
        p = 0.5
        dist = lambda x,y: p
        Y = None
        sampler = abcpmc.Sampler(1, Y, postfn, dist)
        
        
        thetas = np.array([[1], [1], [2]])
        dists = np.array([1, 1, 2])
        ws = np.array([1, 1, 1])
        pool = abcpmc.sampler.PoolSpec(1, eps, 1, thetas, dists, ws)
        
        wrapper = abcpmc.sampler.OLCMParticleProposal(sampler, eps, pool, {})
        
        sigma = np.var(thetas[:1])
        assert wrapper._get_sigma(thetas[0]) == sigma

    def test_get_sigma_multidist(self):
        eps = 1.
        threshold = [eps, eps]

        thetai = 1
        postfn = lambda theta: thetai
        p = 0.5
        dist = lambda x,y: p
        Y = None
        sampler = abcpmc.Sampler(1, Y, postfn, dist)
        
        
        dists = np.array([[eps/2, eps/2], 
                          [eps/2, eps/2], 
                          [eps*2, eps*2], 
                          [eps/2, eps*2], 
                          [eps*2, eps/2] ])
        
        thetas = np.array([[1], [1], [2], [2], [2]])
        ws = np.array([1]*len(dists))
        pool = abcpmc.sampler.PoolSpec(1, threshold, 1, thetas, dists, ws)
        
        wrapper = abcpmc.sampler.OLCMParticleProposal(sampler, threshold, pool, {})
        
        sigma = np.var(thetas[:1])
        assert wrapper._get_sigma(thetas[0]) == sigma

class TestKNNParticleProposal(object):
    
    def test_propose(self):
        eps = 1
        thetai = 1
        postfn = lambda theta: thetai
        p = 0.5
        dist = lambda x,y: p
        Y = None
        sampler = abcpmc.Sampler(1, Y, postfn, dist)
        sigma = 0
        
        thetas = np.array([[1], [1], [2]])
        dists = np.array([1, 1, 2])
        ws = np.array([1, 1, 1])
        pool = abcpmc.sampler.PoolSpec(1, eps, 1, thetas, dists, ws)
        
        wrapper = abcpmc.sampler.KNNParticleProposal(sampler, eps, pool, {})
        
        assert wrapper._get_sigma(thetas[0], 2) == sigma

class TestSampler(object):
    
    def test_sample(self):
        N = 10
        T = 2
        postfn = lambda theta: None
        
        dist = lambda X, Y: 0
        prior = abcpmc.TophatPrior([0], [100])
        sampler = abcpmc.Sampler(N, 0, postfn, dist)
        
        eps_proposal = abcpmc.ConstEps(T, 10)
        for i, pool in enumerate(sampler.sample(prior, eps_proposal)):
            assert pool is not None
            assert len(pool.thetas) == N
        
        assert i+1 == T

def test_weighted_avg_and_std():
    values = np.random.normal(size=1000)
    weights = np.ones((1000))
    
    avg, std = abcpmc.weighted_avg_and_std(values, weights)
    
    assert np.allclose(avg, np.average(values))
    assert np.allclose(std, np.std(values))
    



#!/usr/bin/env python

import os
import sys
from setuptools.command.test import test as TestCommand
from setuptools import find_packages

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
#         self.test_args
#         self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


readme = open('README.rst').read()

history = open('HISTORY.rst').read().replace('.. :changelog:', '')

#during runtime
requires = ["numpy", 
            "scipy>=0.15"]

#for testing
tests_require=['pytest>=2.3', 
               'mock'] 

PACKAGE_PATH = os.path.abspath(os.path.join(__file__, os.pardir))

setup(
    name='abcpmc',
    version='0.1.2',
    description='approximate bayesian computing with population monte carlo',
    long_description=readme + '\n\n' + history,
    author='Joel Akeret',
    author_email='jakeret@phys.ethz.ch',
    url='http://www.cosmology.ethz.ch/research/software-lab/abcpmc.html',
    packages=find_packages(PACKAGE_PATH, "test"),
    package_dir={'abcpmc': 'abcpmc'},
    include_package_data=True,
    install_requires=requires,
    license="GPLv3",
    zip_safe=False,
    keywords=["abcpmc", 
              "approximate bayesian computing ", 
              "population monte carlo"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        'Intended Audience :: Developers',
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Astronomy",      
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
    ],
    tests_require=tests_require,
    cmdclass = {'test': PyTest},
)


# Copyright (C) 2014 ETH Zurich, Institute for Astronomy

'''
Created on Jan 19, 2015

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals
import abcpmc
import numpy as np
    
def test_ConstEps():
    eps_val = 0.5
    T = 5
    eps = abcpmc.ConstEps(T, eps_val)
    
    for i, e in enumerate(eps):
        assert e == eps_val
        
    assert i+1 == T
    
def test_ListEps():
    eps_vals = np.arange(0, 5, 1)
    T = 5 
    eps = abcpmc.ListEps(T, eps_vals)
    
    for e1,e2 in zip(eps, eps_vals):
        assert e1 == e2
    
    assert e1 == eps_vals[-1]
    
def test_LinearEps():
    eps_vals = np.arange(0, 5, 1)
    T = 5 
    eps = abcpmc.LinearEps(T, eps_vals[0], eps_vals[-1])
    
    for e1,e2 in zip(eps, eps_vals):
        assert e1 == e2
    
    assert e1 == eps_vals[-1]
    
def test_LinearConstEps():
    T1 = 6
    T2 = 2
    eps_vals = np.r_[np.linspace(10, 5, T1), [5]*T2]
    eps = abcpmc.LinearConstEps(eps_vals[0], eps_vals[-1], T1, T2)
    
    for e1,e2 in zip(eps, eps_vals):
        assert e1 == e2
    
    assert e1 == eps_vals[-1]
    
def test_ExponentialEps():
    T = 5 
    eps_vals = np.logspace(np.log10(10), np.log10(5), T)
    eps = abcpmc.ExponentialEps(T, eps_vals[0], eps_vals[-1])
    
    for e1,e2 in zip(eps, eps_vals):
        assert e1 == e2
    
    assert e1 == eps_vals[-1]

def test_ExponentialConstEps():
    T1 = 6
    T2 = 2
    min = 5
    max = 10
    eps_vals = np.r_[np.logspace(np.log10(max), np.log10(min), T1), [min]*T2]
    eps = abcpmc.ExponentialConstEps(eps_vals[0], eps_vals[-1], T1, T2)
    
    for e1,e2 in zip(eps, eps_vals):
        assert e1 == e2


# abcpmc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# abcpmc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with abcpmc.  If not, see <http://www.gnu.org/licenses/>.


'''
Created on Oct 9, 2014

author: jakeret
'''
from __future__ import print_function, division, absolute_import, unicode_literals

from multiprocessing.pool import Pool
from collections import namedtuple

import numpy as np
from scipy import stats
from scipy import spatial

__all__ = ["GaussianPrior", 
           "TophatPrior", 
           "ParticleProposal", 
           "KNNParticleProposal", 
           "OLCMParticleProposal", 
           "Sampler", 
           "weighted_cov", 
           "weighted_avg_and_std"
           ]

class GaussianPrior(object):
    """
    Normal gaussian prior
     
    :param mu: scalar or vector of means
    :param sigma: scalar variance or covariance matrix
    """
    
    def __init__(self, mu, sigma):
        self.mu = mu
        self.sigma = sigma
        
    def __call__(self, theta=None):
        if theta is None:
            return np.random.multivariate_normal(self.mu, self.sigma)
        else:
            return stats.multivariate_normal.pdf(theta, self.mu, self.sigma)


class TophatPrior(object):
    """
    Tophat prior
    
    :param min: scalar or array of min values
    :param max: scalar or array of max values
    """
    
    def __init__(self, min, max):
        self.min = np.atleast_1d(min)
        self.max = np.atleast_1d(max)
        assert self.min.shape == self.max.shape
        assert np.all(self.min < self.max)
        
    def __call__(self, theta=None):
        if theta is None:
            return np.array([np.random.uniform(mi, ma) for (mi, ma) in zip(self.min, self.max)])
        else:
            return 1 if np.all(theta < self.max) and np.all(theta >= self.min) else 0

class ParticleProposal(object):
    """
    Creates new particles using twice the weighted covariance matrix (Beaumont et al. 2009)
    """
    def __init__(self, sampler, eps, pool, kwargs):
        self.postfn = sampler.postfn
        self.distfn = sampler.dist
        self.Y = sampler.Y
        self.N = sampler.N
        self.eps = np.asanyarray(eps)
        self.pool = pool
        self.kwargs = kwargs
        
        self.sigma = 2 * weighted_cov(pool.thetas, pool.ws)
    
    def __call__(self, i):
        cnt = 1
        while True:
            idx = np.random.choice(range(self.N), 1, p= self.pool.ws/np.sum(self.pool.ws))[0]
            theta = self.pool.thetas[idx]
            sigma = self._get_sigma(theta, **self.kwargs)
            sigma = np.atleast_2d(sigma)
            thetap = np.random.multivariate_normal(theta, sigma)
            X = self.postfn(thetap)
            p = np.asarray(self.distfn(X, self.Y))
            
            if np.all(p <= self.eps):
                break
            cnt+=1
        return thetap, p, cnt

    def _get_sigma(self, theta):
        return self.sigma

class KNNParticleProposal(ParticleProposal):
    """
    Creates new particles using a covariance matrix from the K-nearest neighbours  (Fillipi et al. 2012)
    Set `k` as key-word arguement in `abcpmc.Sampler.particle_proposal_kwargs`
    """
    
    def _get_sigma(self, theta, k):
        tree = spatial.cKDTree(self.pool.thetas)
        _, idxs = tree.query(theta, k, p=2)
        sigma = np.cov(self.pool.thetas[idxs].T)
        return sigma

class OLCMParticleProposal(ParticleProposal):
    """
    Creates new particles using an optimal loacl covariance matrix (Fillipi et al. 2012)
    """
    
    def _get_sigma(self, theta):
        if len(self.eps.shape) == 0:
            idx = self.pool.dists < self.eps
        else:
            idx = np.all(self.pool.dists < self.eps, axis=1)
        thetas = self.pool.thetas[idx]
        weights = self.pool.ws[idx]
        weights = weights/np.sum(weights)
        
        m = np.sum((weights * thetas.T).T, axis=0)
        n = thetas.shape[1]
        
        sigma = np.empty((n, n))
        for i in range(n):
            for j in range(n):
                sigma[i, j] = np.sum(weights * (thetas[:, i] - m[i]) * (thetas[:, j] - m[j]).T)  + (m[i] - theta[i]) * (m[j] - theta[j])
        return sigma


"""Namedtuple representing a pool of one sampling iteration"""
PoolSpec = namedtuple("PoolSpec", ["t", "eps", "ratio", "thetas", "dists", "ws"])

class Sampler(object):
    """
    ABC population monte carlo sampler
    
    :param N: number of particles
    :param Y: observed data set
    :param postfn: model function (a callable), which creates a new dataset x for a given theta
    :param dist: distance function rho(X, Y) (a callable)
    :param threads: (optional) number of threads. If >1 and no pool is given <threads> multiprocesses will be started
    :param pool: (optional) a pool instance which has a <map> function 
    """
    
    particle_proposal_cls = ParticleProposal
    particle_proposal_kwargs = {}
    
    def __init__(self, N, Y, postfn, dist, threads=1, pool=None):
        self.N = N
        self.Y = Y
        self.postfn = postfn
        self.dist = dist

        if pool is not None:
            self.pool = pool
            self.mapFunc  = self.pool.map
            
        elif threads == 1:
            self.mapFunc = map
        else:
            self.pool = Pool(threads)
            self.mapFunc  = self.pool.map
            
    
    def sample(self, prior, eps_proposal):
        """
        Launches the sampling process. Yields the intermediate results per iteration.
        
        :param prior: instance of a prior definition (or an other callable)  see :py:class:`sampler.GaussianPrior`
        :param eps_proposal: an instance of a threshold proposal (or an other callable) see :py:class:`sampler.ConstEps`
        
        :yields pool: yields a namedtuple representing the values of one iteration
        """
        
        eps = eps_proposal.next()
        wrapper = _RejectionSamplingWrapper(eps, prior, self.postfn, self.dist, self.Y)
        
        res = self.mapFunc(wrapper, range(self.N))
        thetas = np.array([theta for (theta, _, _) in res])
        dists = np.array([dist for (_, dist, _) in res])
        cnts = np.sum([cnt for (_, _, cnt) in res])
        ws = np.ones(self.N) / self.N
        
        pool = PoolSpec(0, eps, self.N/cnts, thetas, dists, ws)
        yield pool
        
        for i, eps in enumerate(eps_proposal):
            t = i+1

            particleProposal = self.particle_proposal_cls(self, eps, pool, self.particle_proposal_kwargs)
            
            res = self.mapFunc(particleProposal, range(self.N))
            thetas = np.array([theta for (theta, _, _) in res])
            dists = np.array([dist for (_, dist, _) in res]) 
            cnts = np.sum([cnt for (_, _, cnt) in res])
            
            sigma = 2 * weighted_cov(pool.thetas, pool.ws)
            wrapper = _WeightWrapper(prior, sigma, pool.ws, pool.thetas)
            
            wt = np.array(list(self.mapFunc(wrapper, thetas)))
            ws = wt/np.sum(wt)
            
            pool = PoolSpec(t, eps, self.N/cnts, thetas, dists, ws)
            yield pool
            
            
    def close(self):
        """
        Tries to close the pool (avoid hanging threads)
        """
        if hasattr(self, "pool") and self.pool is not None:
            self.pool.close()

   
class _WeightWrapper(object):  # @DontTrace
    """
    Wraps the computation of new particle weights.
    Allows for pickling the functionality.
    """
    
    def __init__(self, prior, sigma, ws, thetas):
        self.prior = prior
        self.sigma = sigma
        self.ws = ws
        self.thetas = thetas
    
    def __call__(self, theta):
        kernel = stats.multivariate_normal(theta, self.sigma).pdf
        w = self.prior(theta) / np.sum(self.ws * kernel(self.thetas))
        return w
    
class _RejectionSamplingWrapper(object):  # @DontTrace
    """
    Wraps the computation of new particles in the first iteration (simple rejection sampling).
    Allows for pickling the functionality.
    """
    
    def __init__(self, eps, prior, postfn, dist, Y):
        self.eps = np.asarray(eps)
        self.prior = prior
        self.postfn = postfn
        self.dist = dist
        self.Y = Y
    
    def __call__(self, i):
        cnt = 1
        while True:
            thetai = self.prior()
            X = self.postfn(thetai)
            p = np.asarray(self.dist(X, self.Y))
            if np.all(p <= self.eps):
                break
            cnt+=1
        return thetai, p, cnt

def weighted_cov(values, weights):
    """
    Computes a weighted covariance matrix
    
    :param values: the array of values
    :param weights: array of weights for each entry of the values
    
    :returns sigma: the weighted covariance matrix
    """
    
    n = values.shape[1]
    sigma = np.empty((n, n))
    w = weights.sum() / (weights.sum()**2 - (weights**2).sum()) 
    average = np.average(values, axis=0, weights=weights)
    for j in range(n):
        for k in range(n):
            sigma[j, k] = w * np.sum(weights * ((values[:, j] - average[j]) * (values[:, k] - average[k])))
    return sigma


def weighted_avg_and_std(values, weights, axis=None):
    """
    Return the weighted avg and standard deviation.
    
    :param values: Array with the values
    :param weights: Array with the same shape as values containing the weights
    :param axis: (optional) the axis to be used for the computation
    
    :returns avg, sigma: weighted average and standard deviation
    """
    #http://stackoverflow.com/a/2415343/4067032
    avg = np.average(values, weights=weights, axis=axis)
    # Fast and numerically precise
    variance = np.average((values-avg)**2, weights=weights, axis=axis)
    return (avg, np.sqrt(variance))
